diff -u --recursive libdvdcss-1.2.1/src/libdvdcss.c libdvdcss-1.2.1-cache/src/libdvdcss.c
--- libdvdcss-1.2.1/src/libdvdcss.c	Sun Jun  2 16:54:10 2002
+++ libdvdcss-1.2.1-cache/src/libdvdcss.c	Sat Jul  6 15:18:35 2002
@@ -34,6 +34,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
+#include <dirent.h>
 
 #ifdef HAVE_UNISTD_H
 #   include <unistd.h>
@@ -75,6 +76,8 @@
 static int _dvdcss_open  ( dvdcss_handle, char *psz_target );
 static int _dvdcss_close ( dvdcss_handle );
 static int _dvdcss_readv ( dvdcss_handle, struct iovec *p_iovec, int i_blocks );
+static int _dvdcss_cache_init ( dvdcss_handle );
+static int _dvdcss_get_serial ( dvdcss_handle, char *p_serial );
 
 /*****************************************************************************
  * Local prototypes, win32 specific
@@ -150,6 +153,7 @@
 #ifndef WIN32
     char *psz_raw_device = getenv( "DVDCSS_RAW_DEVICE" );
 #endif
+    char *psz_use_cache = getenv( "DVDCSS_USE_CACHE" );
 
     dvdcss_handle dvdcss;
 
@@ -167,6 +171,8 @@
     dvdcss->i_method = DVDCSS_METHOD_KEY;
     dvdcss->b_debug = 0;
     dvdcss->b_errors = 1;
+    dvdcss->b_cacheloaded = 0;
+    dvdcss->i_cachefd = -1;
 
     /* Find method from DVDCSS_METHOD environment variable */
     if( psz_method != NULL )
@@ -214,6 +220,41 @@
         }
     }
 
+    /* Check cache directory exists */
+    if ( psz_use_cache != NULL )
+    {
+        char p_dname[2048];
+        DIR *cache_dir;
+
+        switch(psz_use_cache[0])
+        {
+        case '1':
+        case 't': 
+        case 'T':
+        case 'Y':
+        case 'y':
+            _dvdcss_debug( dvdcss, "CSS cache usage requested" );
+
+            snprintf(p_dname, sizeof(p_dname), "%s/%s", getenv("HOME"), DVDCSS_CACHE_DIR);
+            mkdir(p_dname, 0777);
+
+            cache_dir = opendir(p_dname);
+            if( cache_dir == NULL )
+            {
+                _dvdcss_error( dvdcss, "unable to open cache directory" );
+                psz_use_cache = NULL;
+            } else {
+                closedir( cache_dir );
+            }
+            break;
+
+        default:
+            _dvdcss_debug( dvdcss, "CSS cache disabled" );
+            psz_use_cache = NULL;
+            break;
+        }
+    }
+
     /* Open device */
     i_ret = _dvdcss_open( dvdcss, psz_target );
     if( i_ret < 0 )
@@ -252,6 +293,16 @@
         }
     }
 
+    /* If disc is encrypted, load any cached keys we know */
+    if ( dvdcss->b_encrypted && (psz_use_cache > 0)) {
+      if( _dvdcss_cache_init( dvdcss ) <= 0 ) {
+	_dvdcss_close( dvdcss );
+	free( dvdcss );
+	return NULL;
+      }
+      _dvdcss_debug(dvdcss, "Cache initialised");
+    }
+
 #ifndef WIN32
     if( psz_raw_device != NULL )
     {
@@ -300,14 +351,21 @@
 {
     dvd_title_t *p_title;
     dvd_title_t *p_newtitle;
     dvd_key_t    p_title_key;
     int          i_ret;
+    int                i;
+    FILE        *p_fd;
 
     if( ! dvdcss->b_encrypted )
     {
         return 0;
     }
 
+    if (( dvdcss->i_cachefd > -1 ) && ( ! dvdcss->b_cacheloaded ))
+    {
+        _dvdcss_error(dvdcss, "Cache in use, but not primed!");
+    }
+
     /* Check if we've already cracked this key */
     p_title = dvdcss->p_titles;
     while( p_title != NULL
@@ -322,6 +381,8 @@
     {
         /* We've already cracked this key, nothing to do */
         memcpy( dvdcss->css.p_title_key, p_title->p_key, sizeof(dvd_key_t) );
+        _dvdcss_debug(dvdcss, "found key in cache");
+
         return 0;
     }
 
@@ -335,6 +396,6 @@
     }
     else if( i_ret == 0 )
     {
-        _dvdcss_debug( dvdcss, "unecrypted title" );
+        _dvdcss_debug( dvdcss, "unencrypted title" );
         /* Still store this in the cache, so we don't need to check again. */
     }
@@ -369,6 +437,37 @@
     }
 
     memcpy( dvdcss->css.p_title_key, p_title_key, KEY_SIZE );
+
+    _dvdcss_debug( dvdcss, "Added key to memory cache" );
+
+    if( dvdcss->i_cachefd > -1 )
+    {
+      do
+      {
+        /* Add to cache file */
+        _dvdcss_debug( dvdcss, "Adding key to disk cache" );
+
+	p_fd = fdopen(dvdcss->i_cachefd, "a+");
+	if(p_fd == NULL) {
+	  _dvdcss_error(dvdcss, "Could not get file descriptor for cache!");
+	  break;
+	}
+
+        fseek(p_fd, 0, SEEK_END);
+
+        fprintf(p_fd, "%ld:", i_block);
+        for( i = 0; i < KEY_SIZE; i++)
+        {
+            if( i > 0 )
+                fprintf(p_fd, ", ");
+
+            fprintf(p_fd, "0x%x", p_title_key[i]);
+        }
+        fprintf(p_fd, "\n");
+	fflush(p_fd);
+      } while(0);
+    }
+
     return 0;
 }
 
@@ -655,6 +754,8 @@
 
 #else
     close( dvdcss->i_fd );
+    close(dvdcss->i_cachefd);
+    dvdcss->i_cachefd = -1;
 
     if( dvdcss->i_raw_fd >= 0 )
     {
@@ -778,6 +879,159 @@
 #endif
 }
 
+static int _dvdcss_cache_init( dvdcss_handle dvdcss )
+{
+    int i, b_addkey;
+    int i_ret;
+    char *p_cur, *p_token;
+    char p_serial[SERIAL_SIZE + 1];
+    char p_buf[2048];
+    char psz_debug[2048];
+    dvd_title_t curtitle;
+    dvd_title_t *p_title;
+    dvd_title_t *p_newtitle;
+    FILE *p_fd;
+
+    _dvdcss_debug(dvdcss, "Using CSS cache");
+    
+    if (_dvdcss_get_serial(dvdcss, p_serial) <= 0)
+    {
+	_dvdcss_error(dvdcss, "Error reading serial number");
+	return 0;
+    }
+    
+    /* Create/open cache file */
+    snprintf(p_buf, sizeof(p_buf), "%s/%s/%s", getenv("HOME"), DVDCSS_CACHE_DIR, p_serial);
+    dvdcss->i_cachefd = open(p_buf, (O_RDWR | O_CREAT), 0666);
+    
+    /* Get a file descriptor */
+    p_fd = fdopen(dvdcss->i_cachefd, "r+");
+    if (p_fd == NULL)
+    {
+      _dvdcss_error(dvdcss, "Could not get stream for CSS cache!");
+      return 0;
+    }
+    
+    /* Prime the title key cache */
+    while( (p_fd != NULL) && (fgets(p_buf, sizeof(p_buf), p_fd) != NULL) )
+    {
+	/* Parse line */
+	curtitle.i_startlb = 0;
+	memset(curtitle.p_key, 0, KEY_SIZE);
+	
+	p_cur = p_buf;
+	
+	do
+	{
+	    /* Line format
+	       block:key1,key2,...
+	    */
+	    b_addkey = 0;
+	    
+	    p_token = strsep(&p_cur, ":");
+	    if( p_token == NULL)
+	      break;
+	    
+	    curtitle.i_startlb = strtol(p_token, (char **)NULL, 10);
+	    
+	    p_token = strsep(&p_cur, ":");
+	    if( p_token == NULL)
+	      break;
+	    
+	    p_cur = p_token;
+	    for(i = 0; i < KEY_SIZE; i++)
+            {
+		p_token = strsep(&p_cur, ",");
+		if (p_token == NULL)
+		  break;
+		
+		curtitle.p_key[i] = strtol(p_token, (char **)NULL, 0);
+            }
+	    
+	    b_addkey = 1;
+        } while(0);
+	
+	/* If we have a valid key, add it to the cache list */
+	if( b_addkey > 0 )
+        {
+	    snprintf(psz_debug, sizeof(psz_debug), "Read cached key for block %ld", curtitle.i_startlb);
+	    _dvdcss_debug(dvdcss, psz_debug);
+	    
+	    p_title = dvdcss->p_titles;
+	    p_newtitle = NULL;
+	    while( (p_title != NULL) && (p_title->i_startlb < curtitle.i_startlb) )
+            {
+		p_newtitle = p_title;
+		p_title = p_title->p_next;
+            }
+	    
+	    p_title = p_newtitle;
+	    p_newtitle = malloc(sizeof(dvd_title_t));
+	    p_newtitle->i_startlb = curtitle.i_startlb;
+	    memcpy( p_newtitle->p_key, curtitle.p_key, KEY_SIZE );
+	    
+	    /* Link the new title, either at the beginning or inside the list */
+	    if( p_title == NULL )
+            {
+		p_newtitle->p_next = dvdcss->p_titles;
+		dvdcss->p_titles = p_newtitle;
+                p_newtitle->p_titles = NULL;
+            }
+	    else
+            {
+		p_newtitle->p_next = p_title->p_next;
+		p_title->p_next = p_newtitle;
+            }
+	    
+	    
+        }
+	else
+	{
+	    snprintf(psz_debug, sizeof(psz_debug), "Read invalid CSS cache line : %s", p_buf);
+	    _dvdcss_error(dvdcss, psz_debug);
+        }
+    }
+    
+    dvdcss->b_cacheloaded = 1;
+    return 1;
+}
+
+static int _dvdcss_get_serial( dvdcss_handle dvdcss, char *p_serial )
+{
+    int i;
+    int i_ret, i_seekpos;
+    char p_block[DVDCSS_BLOCK_SIZE];
+    char psz_debug[2048];
+
+    i_seekpos = dvdcss->i_seekpos;
+    i_ret = _dvdcss_seek( dvdcss, 32 );
+    if( i_ret < 0 )
+    {
+	_dvdcss_error(dvdcss, "Error seeking to serial block");
+	return 0;
+    }
+    
+    i_ret = _dvdcss_read( dvdcss, p_block, 1);
+    if( i_ret < 0 )
+    {
+	_dvdcss_error(dvdcss, "Error reading serial block");
+	return 0;
+    }
+    
+    _dvdcss_seek( dvdcss, i_seekpos );
+    
+    memset(p_serial, 0, (SERIAL_SIZE + 1));
+    for(i = 73; i < 89; i++)
+    {
+      if( (p_block[i] > 32) && (p_block[i] < 127))
+	  p_serial[i - 73] = p_block[i];
+    }
+    
+    snprintf(psz_debug, sizeof(psz_debug), "Got DVD serial number: %s", p_serial);
+    _dvdcss_debug(dvdcss, psz_debug);
+    
+    return 1;
+}
 
 #if defined( WIN32 )
 
diff -u --recursive libdvdcss-1.2.1/src/libdvdcss.h libdvdcss-1.2.1-cache/src/libdvdcss.h
--- libdvdcss-1.2.1/src/libdvdcss.h	Thu Apr  4 15:21:25 2002
+++ libdvdcss-1.2.1-cache/src/libdvdcss.h	Wed Jun 26 21:09:04 2002
@@ -31,6 +31,10 @@
     int i_fd;
     int i_seekpos;
 
+    /* Cache data */
+    int i_cachefd;
+    boolean_t b_cacheloaded;
+
     /* Decryption stuff */
     int          i_method;
     css_t        css;
@@ -58,6 +62,12 @@
 #define DVDCSS_METHOD_KEY        0
 #define DVDCSS_METHOD_DISC       1
 #define DVDCSS_METHOD_TITLE      2
+
+/*****************************************************************************
+ * CSS key cache
+ *****************************************************************************/
+#define SERIAL_SIZE		16
+#define DVDCSS_CACHE_DIR	".dvdcss"
 
 /*****************************************************************************
  * Functions used across the library
